# SUZUKI Tetsuya <suzuki@spice-of-life.net>
# $Id: tml-tapkit.rb,v 1.21 2004/07/05 08:37:40 john Exp $

require 'tapkit'

include TapKit

module SharingContext
  def app()
    TML::TapKit.app
  end

  def ec()
    unless @ec then
      @ec = TML::TapKit.app.ec
    end
    @ec
  end

  def entity()
    unless @entity then
      @entity = app.entity entity_name
    end
    @entity
  end

  def create
    object = ec.create entity_name
    set_default object
    object
  end

  # ֥Ȥ˥ǥեͤ򥻥å
  def set_default(object); end

  def fetchspec(k)
    q  = Qualifier.format("%K == %@", [key, k])
    fs = FetchSpec.new(entity_name, q)
    fs
  end

  def ml_object(ml)
    q  = Qualifier.format("name == %@", [ml])
    fs = FetchSpec.new('ML', q)
    fs.limit = 1
    ec.fetch(fs).first
  end
end

class TML; end

class TML::TapKit
  @@app = nil

  MODEL = "tapkit.yaml"

  def self.app()
    unless @@app then
      unless $tapkit_model then $tapkit_model = "#$tml_dir/#{MODEL}" end
      # @@app = Application.new($tapkit_model, '/tmp')
      @@app = Application.new($tapkit_model)
      @@app.log_options[:sql] = $tapkit_log_sql
      if $tapkit_log_file then
        @@app.log_options[:out] = File.new($tapkit_log_file, 'a+')
      end
    end
    @@app
  end

  class Lock
    include SharingContext

    @@locks = []

    def fetchspec(ident)
      q  = Qualifier.format("ident == %@", [ident])
      fs = FetchSpec.new('Lock', q)
      fs
    end

    def self.callback(ident)
      proc do
        fs = fetchspec ident
        objects = ec.fetch fs
        objects.each do |obj|
          ec.delete obj
        end
        ec.save

        TML.log "unlock: #{ident}"
        @@locks.delete ident
      end
    end

    def initialize(ident)
      @ec = app.ec

      if @@locks.include? ident then
        raise "dead lock detected: #{ident}"
      end

      lock ident

      ObjectSpace.define_finalizer(self, Lock.callback(ident))
      @ident = ident
      @@locks << ident
      TML.log "lock: #{ident}"
    end

    def lock(ident)
      fs = fetchspec ident
      fs.limit = 1
      obj = ec.fetch(fs).first

      st = Time.now
      begin
        if obj and \
          obj['timestamp'].to_time.to_i < Time.now.to_i-TML.param(:max_lock_time) then
          ec.delete obj
          ec.save
          TML.log "#{ident}: max_lock_time(#{TML.param(:max_lock_time)}) exceeded"
        end
        newobj = ec.create 'Lock'
        newobj['ident'] = ident
        ec.save
      rescue
        if st+TML.param(:lock_timeout) < Time.now then
          raise Error.new(:ETIMEDOUT, "lock timed out")
        end
        sleep 1
        retry
      end
    end

    def unlock()
      fs = fetchspec @ident
      objects = ec.fetch fs
      objects.each do |obj|
        ec.delete obj
      end
      ec.save

      TML.log "unlock: #{@ident}"
      ObjectSpace.undefine_finalizer(self)
      @@locks.delete @ident
    end
  end

  class Meta
    include SharingContext

    private

    def alt_key() nil end
    def log_str() entity_name end

    def check_attr(attr)
      keys = app.entity(entity_name).class_description.attribute_keys
      attr.keys.each do |k|
        unless keys.include? k then
          TML::Error.new(:EINVAL, "#{attr}: no such attribute")
        end
      end
    end

    def qualifier(k)
      unless alt_key then
        Qualifier.format("%K == %@", [key, k])
      else
        Qualifier.format("(%K == %@) and (%K == %@)", [key, k, alt_key, @alt_value])
      end
    end

    public

    def have?(k)
      q = qualifier k
      fs = FetchSpec.new(entity_name, q)
      objects = ec.fetch fs

      if objects.empty? then
        false
      else
        true
      end
    end

    def list()
      if alt_key then
        q = Qualifier.format("%K = %@", [alt_key, @alt_value])
      else
        q = nil
      end
      fs = FetchSpec.new(entity_name, q)
      objects = ec.fetch fs

      result = []
      objects.each do |obj|
        result << obj[key]
      end
      result
    end

    def list_with_data()
      if alt_key then
        q = Qualifier.format("%K = %@", [alt_key, @alt_value])
      else
        q = nil
      end
      fs = FetchSpec.new(entity_name, q)
      objects = ec.fetch fs

      result = []
      objects.each do |obj|
        data = {}
        obj.attribute_keys.each do |key|
          if items.include? key then
            data[key] = obj[key]
          end
        end
        result << data
      end
      result
    end

    def add(k, attr={})
      check_attr attr
      TML.log "add #{log_str}: #{k}" if log_str
    end

    # TapKitϡDBͭΥ顼Ϥ狼ʤʥåǳǧǤΤߡ
    # ߥåȤ˼Ԥ饨顼Ϥ
    def delete(k)
      fs = fetchspec k
      fs.limit = 1
      obj = ec.fetch(fs).first
      if obj then
        ec.delete obj
        ec.save
      else
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end

      TML.log "delete #{log_str}: #{k}" if log_str
    end

    def set(k, attr)
      unless have? k then
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end

      fs = fetchspec k
      fs.limit = 1
      obj = ec.fetch(fs).first
      attr.each do |key, value|
        obj[key] = value
      end
      ec.save

      TML.log "set #{log_str}: #{k}" if log_str
    end

    def get(k)
      unless have? k then
        raise TML::Error.new(:ENOENT, "#{k}: not found")
      end

      data = {}
      fs = fetchspec k
      fs.limit = 1
      obj = ec.fetch(fs).first
      obj.attribute_keys.each do |key|
        if items.include? key then
          data[key] = obj[key]
        end
      end
      data
    end
  end

  class ML < Meta
    def entity_name() "ML" end
    def key() "name" end
    def items()
      %w(sender overview not_member subject_mlname
        counter_width remote_command remote_admin spool web_interface
        need_confirm_command allowed_command anonymous_command admin_command
        logging replyto_type
      )
    end
    def log_str() "ml" end

    def set_default(object)
      object['counter_width']  = 5
      object['not_member']     = false
      object['subject_mlname'] = true
      object['remote_command'] = false
      object['remote_admin']   = false
      object['spool']          = true
      object['web_interface']  = true
      object['logging']        = true
    end

    def add(k, attr={})
      if have?(k) then
        raise TML::Error.new(:EEXIST, "#{k}: already exist")
      end

      obj = create
      obj['name'] = k
      ec.save
      super
    end
  end

  class Counter < Meta
    def entity_name() "Counter" end
    def key() "ml.name" end
    def items() ["counter"] end
    def log_str() "counter" end

    def add(ml)
      mlobj = ml_object ml
      obj = create
      obj['ml'] = mlobj
      obj['counter'] = 0
      ec.save
      super
    end

    def set(ml, cnt)
      super(ml, {"counter"=>cnt.to_s})
    end

    def get(ml)
      super(ml)["counter"].to_i
    end
  end

  class Passwd < Meta
    def entity_name() "Address" end
    def key() "address" end
    def items() ["passwd"] end
    def log_str() "password" end

    def add(addr, pw)
      if have?(addr) then
        raise TML::Error.new(:EEXIST, "#{addr}: already exist")
      end

      obj = create
      obj['address']  = addr
      obj['password'] = pw
      ec.save
      super(addr, {"password"=>pw})
    end

    def set(addr, pw)
      super(addr, {"password"=>pw})
    end

    def auth(addr, pw)
      begin
        p = get(addr)["password"]
        if p == pw then
          return true
        end
      rescue TML::Error => err
        raise unless err.id == :ENOENT
      end
      TML.log "auth failed: #{addr}"
      return false
    end
  end

  class MetaML < Meta
    def alt_key() "ml.name" end

    def initialize(ml)
      @alt_value = ml
      @mlobj = ml_object ml
    end

    def delete_all()
      q = Qualifier.format("%K == %@", [alt_key, @alt_value])
      fs = FetchSpec.new(entity_name, q)
      ec.fetch(fs).each do |obj|
        ec.delete obj
      end
      ec.save
    end

    def address_object(addr)
      q = Qualifier.format('address == %@', [addr])
      fs = FetchSpec.new('Address', q)
      fs.limit = 1

      # ɥ쥹ʤФǺ
      unless obj = ec.fetch(fs).first then
        obj = ec.create 'Address'
        obj['address'] = addr
        ec.save
      end
      obj
    end

    # ɥ쥹˴Ϣơ֥ Member, Admin 줿顢
    # ɥ쥹
    def delete_address(k)
      q = Qualifier.format("address.address == %@", [k])
      fs = FetchSpec.new('Member', q)
      members = ec.fetch(fs)
      fs = FetchSpec.new('Admin', q)
      admins = ec.fetch(fs)

      if members.empty? and admins.empty? then
        q = Qualifier.format("address == %@", [k])
        fs = FetchSpec.new('Address', q)
        ec.fetch(fs).each { |obj| ec.delete obj }
        ec.save
      end
    end
  end

  class Member < MetaML
    def entity_name() "Member" end
    def key() "address.address" end
    def items() [] end
    def log_str() "member" end

    def add(addr, attr={})
      if have?(addr) then
        raise TML::Error.new(:EEXIST, "#{addr}: already exist")
      end

      obj = create
      obj['address']  = address_object addr
      obj['ml'] = @mlobj 
      ec.save
      super
    end

    def delete(k)
      super
      delete_address k
    end
  end

  class Admin < MetaML
    def entity_name() "Admin" end
    def key() "address.address" end
    def items() [] end
    def log_str() "admin" end

    def add(addr, attr={})
      if have?(addr) then
        raise TML::Error.new(:EEXIST, "#{addr}: already exist")
      end

      obj = create
      obj['address']  = address_object addr
      obj['ml'] = @mlobj 
      ec.save
      super
    end

    def delete(k)
      super
      delete_address k
    end
  end

  class Child < MetaML
    def entity_name() "Tree" end
    def key() "child.name" end
    def log_str() "child" end
    def items() [] end
    def alt_key() "parent.name" end

    def add(ml, attr={})
      child = ml_object ml
      obj = create
      obj['parent'] = @mlobj
      obj['child']  = child
      ec.save
      super
    end
  end

  class Parent < MetaML
    def entity_name() "Tree" end
    def key() "parent.name" end
    def log_str() "parent" end
    def items() [] end
    def alt_key() "child.name" end

    def add(ml, attr={})
      parent = ml_object ml
      obj = create
      obj['parent'] = parent
      obj['child'] = @mlobj
      ec.save
      super
    end
  end

  class Spool < MetaML
    def entity_name() "Spool" end
    def key() "seq" end
    def items() ["subject", "body"] end
    def log_str() "spool" end

    def spool(seq, subject, body)
      add(seq, {"subject"=>subject, "body"=>body})
    end

    def add(seq, attr={})
      if have?(seq) then
        raise TML::Error.new(:EEXIST, "#{seq}: already exist")
      end

      obj = create
      obj['ml'] = @mlobj
      obj['seq'] = seq
      obj['subject'] = attr['subject']
      obj['body'] = attr['body']
      ec.save
      super
    end

    def copy_messages(msgs, dest)
      q = or_qualifier('seq', msgs)
      fs = FetchSpec.new('Spool', q)
      objs = ec.fetch fs
      objs.each do |obj|
        File.open("#{dest}/#{obj['seq']}", "w") {|f| f.print obj['body']}
      end
    end

    def summary_list(msgs=nil)
      q = or_qualifier('seq', msgs)
		sort = SortOrdering.new('seq', SortOrdering::ASC)
      fs = FetchSpec.new('Spool', q, [sort])
      objs = ec.fetch fs

      ret = ""
      objs.each do |obj|
        ret << "#{obj['seq']} #{obj['subject']}\n"
      end
      ret
    end

    private

    def or_qualifier(key, array)
      qualifiers = []
      array.each do |item|
        qualifiers << Qualifier.format("%K == %@", [key, item])
      end
      OrQualifier.new qualifiers
    end
  end

  class Work < MetaML
    def entity_name() "Work" end
    def key() "lock.ident" end
    def items() ["data"] end

    def add(key, data)
      q = Qualifier.format("lock.ident == %@", [key])
      fs = FetchSpec.new('Lock', q)
      lockobj = ec.fetch fs

      obj = create
      obj['lock'] = lockobj
      obj['ml'] = @mlobj
      obj['data'] = data
      ec.save
      super(key, {"data"=>data})
    end

    def get(key)
      super(key)["data"]
    end

    def expire()
      q = Qualifier.format("timestamp < %@", [Timestamp.now-24*60*60])
      fs = FetchSpec.new('Lock', q)
      ec.fetch(fs).each { |obj| ec.delete obj }
      ec.save
    end
  end

end


module TapKit
   class Attribute
      def _convert_bool( value )
         case adapter_value_type
         when ADAPTER_NUMBER_TYPE
            case value.to_s
            when 'true'  then 1
            when 'false' then 0
            else              nil
            end
         when ADAPTER_CHARACTERS_TYPE
            bool = value.to_s.downcase
            if (bool != true) and (bool != false) then
               nil
            else
               bool
            end
         else
            raise "can't cast '#@external_type' to boolean"
         end
      end
   end
end


# Test
#
if __FILE__ == $0 then
  require "tml.rb"
  def TML.log(a) puts a end

  $tapkit_model = "tapkit.yaml"
  # TML::TapKit::app.log_options[:sql] = true

  # Lock test
  l = TML::TapKit::Lock.new("hoge")
  begin
    TML::TapKit::Lock.new("hoge")
    raise "error!"
  rescue => e
    raise unless e.to_s == "dead lock detected: hoge"
  end
  l.unlock
  TML::TapKit::Lock.new("hoge").unlock

  # ML test
  ml = TML::TapKit::ML.new
  ml.add("hoge")
  raise "error!" unless ml.list == ["hoge"]
  ml.set("hoge", {"overview"=>"HOGEHOGE", "not_member"=>"true"})
  h = ml.get("hoge")
  raise "error!" if h["overview"] != "HOGEHOGE" or h["not_member"] != "true"

  # Counter test
  ct = TML::TapKit::Counter.new
  ct.add("hoge")
  raise "error!" unless ct.get("hoge") == 0
  ct.set("hoge", 99)
  raise "error!" unless ct.get("hoge") == 99
  ct.delete("hoge")
  raise "error!" if ct.have? "hoge"

  # Passwd test
  pw = TML::TapKit::Passwd.new
  pw.add("hoge", "hogepw")
  raise "error!" unless pw.list == ["hoge"]
  pw.auth("hoge", "hogepw")
  pw.set("hoge", "hogepw2")
  pw.auth("hoge", "hogepw2")
  pw.delete("hoge")
  raise "error!" if pw.have? "hoge"

  # Member test
  mm = TML::TapKit::Member.new("hoge")
  mm.add("hogehoge")
  raise "error!" unless mm.list == ["hogehoge"]
  mm.delete("hogehoge")
  raise "error!" if mm.have? "hogehoge"

  # Admin test
  ad = TML::TapKit::Admin.new("hoge")
  ad.add("xxxx")
  raise "error!" unless ad.list == ["xxxx"]
  ad.delete("xxxx")
  raise "error!" if ad.have? "xxxx"

  # Tree test
  ml.add("fuga")
  ch = TML::TapKit::Child.new("hoge")
  ch.add("fuga")
  raise "error!" unless ch.list == ["fuga"]
  raise "error!" unless TML::TapKit::Parent.new("fuga").list == ["hoge"]
  ch.delete("fuga")
  raise "error!" if ch.have? "fuga"
  ml.delete("fuga")
  ml.delete("hoge")
  raise "error!" if ml.have? "hoge"

  # Spool test
  ml.add "hoge"
  sp = TML::TapKit::Spool.new("hoge")
  sp.spool(1, "subject1", "body1\nbody1\n")
  sp.spool(2, "subject2", "body2\nbody2\n")
  raise "error!" if sp.summary_list([1, 2]) != "1 subject1\n2 subject2\n"
  sp.copy_messages([1], '.')
  raise "error!" unless FileTest.exist? '1'
  File.delete '1'
  sp.delete_all  
  ml.delete "hoge"
end
